//---------------------------------------------------------------------------
#ifndef commonH
#define commonH
//---------------------------------------------------------------------------
#define PROG_NAME                 "Ptp2Wav"
#define PROG_EXE_NAME             "ptp2wav"
#define PROG_VERSION              0x11
#define PROG_AUTHOR               "RM-TEAM 05/2023"

//---------------------------------------------------------------------------
#define BYTE                      unsigned char
#define WORD                      unsigned short int
#define DWORD                     unsigned int
#define BOOL                      int
#define FALSE                     0
#define TRUE                      !FALSE

//---------------------------------------------------------------------------
#define ERR_OK                    0
#define ERR_PRINT_USAGE           -1
#define ERR_PRINT_VERSION         -2
#define ERR_OUT_OF_MEMORY         -3
#define ERR_UNKNOWN_OPTION        -4
#define ERR_BAD_OPTIONS           -5
#define ERR_NO_OPTION_VALUE       -6
#define ERR_NO_INPUT_FILE         -7
#define ERR_INPUT_FILE_MORE       -8
#define ERR_OUTPUT_FILE_MORE      -9
#define ERR_OPEN_SRC_FILE         -10
#define ERR_ACCESS_SRC_FILE       -11
#define ERR_LENGTH_SRC_FILE       -12
#define ERR_READ_SRC_FILE         -13
#define ERR_OPEN_DEST_FILE        -14
#define ERR_ACCESS_DEST_FILE      -15
#define ERR_WRITE_DEST_FILE       -16
#define ERR_BAD_VALUE             -17
#define ERR_TAPE_CORRUPTED        -18
#define ERR_BAD_BLOCK_NUMBER      -19

#define BLOCK_NUMBER_LIST_SIZE    32

//---------------------------------------------------------------------------
#define HI_STATE                  0xDA
#define MI_STATE                  0x80
#define LO_STATE                  (0x100-HI_STATE)

#define SAMPLE_RATE_LOW           0
#define SAMPLE_RATE_MIDDLE        1
#define SAMPLE_RATE_HIGH          2

#define SAMPLE_RATE_LOW_HZ        11025
#define SAMPLE_RATE_MIDDLE_HZ     22050
#define SAMPLE_RATE_HIGH_HZ       44100

#define SILENT_GAP                4               // 4 sec

// PMD 85
#define PULSE_FREQ_1200           1200.0          // T=833.33 us
// hodnoty siriek puzov musia byt parne!!!
#define PULSE_SIZE_1200_L         10              // ((int) ((SAMPLE_RATE_LOW_HZ / PULSE_FREQ_1200) + 0.5))
#define PULSE_SIZE_1200_M         18              // ((int) ((SAMPLE_RATE_MIDDLE_HZ / PULSE_FREQ_1200) + 0.5))
#define PULSE_SIZE_1200_H         36              // ((int) ((SAMPLE_RATE_HIGH_HZ / PULSE_FREQ_1200) + 0.5))
#define PULSE_FREQ_2400           2400.0          // T=416.66 us
#define PULSE_SIZE_2400_L         6               // ((int) ((SAMPLE_RATE_LOW_HZ / PULSE_FREQ_2400) + 0.5))
#define PULSE_SIZE_2400_M         10              // ((int) ((SAMPLE_RATE_MIDDLE_HZ / PULSE_FREQ_2400) + 0.5))
#define PULSE_SIZE_2400_H         18              // ((int) ((SAMPLE_RATE_HIGH_HZ / PULSE_FREQ_2400) + 0.5))

#define HEAD_LEADER_DURATION      2.8             // 2.8 sec
#define HBODY_LEADER_DURATION     0.5             // 0.5 sec
#define BODY_LEADER_DURATION      0.5             // 0.5 sec
#define TAIL_DURATION             0.5             // 0.5 sec

// Mato
#define PULSE_FREQ                3215.4341       // T=311 us
#define PULSE_SIZE_L              ((int) ((SAMPLE_RATE_LOW_HZ / PULSE_FREQ) + 0.5))
#define PULSE_SIZE_M              ((int) ((SAMPLE_RATE_MIDDLE_HZ / PULSE_FREQ) + 0.5))
#define PULSE_SIZE_H              ((int) ((SAMPLE_RATE_HIGH_HZ / PULSE_FREQ) + 0.5))

#define HEAD_LEADER_PULSES        510
#define HBODY_LEADER_PULSES       128
#define BODY_LEADER_PULSES        128

//---------------------------------------------------------------------------
#pragma pack(push, 1)
typedef struct tagTapeBlock {
  BYTE btCislo;
  char cTyp;
  WORD wZac;
  WORD wLen;
  char cMeno[8];
  BYTE btCrc;
  BYTE btBodyCrc;
  long lOffsetHeader;
  long lOffsetBody;
  BOOL bHeadCrcError;
  BOOL bBodyCrcError;
  BOOL bBodyCrcErrorFix;
  BOOL bRawFile;
  long lBodyLengthError;
  struct tagTapeBlock *prev;
  struct tagTapeBlock *next;
} TAPE_BLOCK;
#pragma pack(pop)

//---------------------------------------------------------------------------
#ifndef MAKEFOURCC
typedef DWORD   FOURCC;

#define MAKEFOURCC(ch0, ch1, ch2, ch3)  ((DWORD)(BYTE)(ch0) \
                                         | ((DWORD)(BYTE)(ch1) << 8) \
                                         | ((DWORD)(BYTE)(ch2) << 16) \
                                         | ((DWORD)(BYTE)(ch3) << 24 ))
#endif

//---------------------------------------------------------------------------
#pragma pack(push, 1)

#ifndef WAVE_FORMAT_PCM
// flags for wFormatTag field of WAVEFORMAT
#define WAVE_FORMAT_PCM     1

// OLD general waveform format structure (information common to all formats)
typedef struct {
  WORD wFormatTag;        // format type
  WORD nChannels;         // number of channels (i.e. mono, stereo, etc.)
  DWORD nSamplesPerSec;   // sample rate
  DWORD nAvgBytesPerSec;  // for buffer estimation
  WORD nBlockAlign;       // block size of data
} WAVEFORMAT;

// specific waveform format structure for PCM data
typedef struct {
  WAVEFORMAT wf;
  WORD wBitsPerSample;
} PCMWAVEFORMAT;
#endif

//---------------------------------------------------------------------------
typedef struct {
  FOURCC riff;
  DWORD dwTotLength;

  FOURCC wave;
  FOURCC fmt;
  DWORD dwHdrLength;

  PCMWAVEFORMAT waveFormat;

  FOURCC data;
  DWORD dwDataLength;
} WAVE_HEADER;

#pragma pack(pop)

//---------------------------------------------------------------------------
typedef struct {
  char *pFnIn;          // pointer to the source file path/name
  char *pFnOut;         // pointer to the destination file path/name
  char szFn[256];       // destination file path/name
  BYTE *pbtBuffer;      // pointer na pracovny buffer
  BYTE *pbtLevels;      // pointer na buffer pre sekvencie logickych urovni
  BYTE *pbtLog0;        // pointer na Log0 sekvenciu
  BYTE *pbtLog1;        // pointer na Log1 sekvenciu
  BOOL bVB;             // verbose print
  BOOL bMato;           // protokol Mato
  BOOL bList;           // zobrazenie obsahu pasky
  BOOL bRevPhase;       // obratena faza signalu
  BOOL bDoubleSpeed;    // dvojnasobna rychlost 2400Bd (iba pre PMD 85)
  int iLog0Length;      // dlzka Log0 sekvencie
  int iLog1Length;      // dlzka Log1 sekvencie
  int iSampleRateType;  // typ vzorkovacej frekvencie
  int iSampleRate;      // vzorkovacia frekvencia
  int iBlockCount;      // pocet blokov na paske
  int *piBlkNumList;    // pointer na zoznam cisel blokov
  int iBlkListCnt;      // pocet cisel blokovv zozname
  int iSilentGap;       // medzera pred dalsim hlavickovym blokom v sekundach
  BYTE btAmpHi;         // horna amplituda impulzov
  BYTE btAmpLo;         // dolna amplituda impulzov
  TAPE_BLOCK *pBlocks;  // pointer na zoznam blokov
  WAVE_HEADER *pWH;     // pointer na WAVE_HEADER strukturu
  FILE *fOut;           // pointer na otvoreny vystupny subor
} WORK_PARAMS;

//---------------------------------------------------------------------------
// main.c
long setFilePointer(FILE *f, long lOffset, int iWhence);
int readFile(char *pszFileName, BYTE *pbtBuff, long lOffset, DWORD dwLength);
int allocBuffer(void **ppvBuff, int iSize);
int reallocBuffer(void **ppvBuff, int iNewSize, int iOldSize);
int findInList(int *piList, int iListSize, int iVal);
int printError(int iErr, void *pPar, void *pPar2);

// tape.c
BOOL readTapeFile(WORK_PARAMS *pWP);
void printBlockList(WORK_PARAMS *pWP);
int loadTapeBlock(WORK_PARAMS *pWP, int iBlk, BOOL bHead, TAPE_BLOCK **ppBlk);
void freeAllBlocks(TAPE_BLOCK *pBlocks);

// wave.c
int createWav(WORK_PARAMS *pWP);

//---------------------------------------------------------------------------
#endif
//---------------------------------------------------------------------------

