//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "common.h"

//---------------------------------------------------------------------------
int findBlock(WORK_PARAMS *pWP, int iBlk, TAPE_BLOCK **ppBlk);
BOOL checkCrc(BYTE *pbtBuff, int iLength, BYTE *pGoodCrc);
BOOL checkHeader(BYTE *pbtBuff, TAPE_BLOCK *pBlk);

//---------------------------------------------------------------------------
void printBlockList(WORK_PARAMS *pWP)
{
  TAPE_BLOCK *pBlk;
  char szName[9] = { 0 };
  int iErr, iBegin, iLength, iIdx, iBlk;
  BOOL bHCE = FALSE, bBCE = FALSE, bBLE = FALSE;

  printf(" Tape file: %s\n", pWP->pFnIn);
  printf(" Blocks: %d\n\n", pWP->iBlockCount);
  printf(" BLK NR/T Name\t\t    Begin\t   Length\n");


  for (iIdx = 0; iIdx < pWP->iBlkListCnt; iIdx++) {
    iBlk = pWP->piBlkNumList[iIdx];

    iErr = findBlock(pWP, iBlk, &pBlk);
    if (iErr != ERR_OK)
      return;

    iBegin = (int) pBlk->wZac;
    iLength = (int) pBlk->wLen;
    if (pBlk->cTyp == 0)
      printf("% 3d: ...\t\t\t\t%5d (%04Xh)\n", iBlk, iLength, iLength);
    else {
      memcpy(szName, pBlk->cMeno, 8);
      printf("% 3d: %02d/%c %8.8s\t%5d (%04Xh)\t%5d (%04Xh)\t%s %s %s\n",
             iBlk, (int) pBlk->btCislo, pBlk->cTyp, szName,
             iBegin, iBegin, iLength, iLength,
             (pBlk->bHeadCrcError) ? "HCE" : "",
             (pBlk->bBodyCrcError) ? "BCE" : "",
             (pBlk->lBodyLengthError >= 0) ? "BLE" : "");

      if (pBlk->bHeadCrcError)
        bHCE = TRUE;
      if (pBlk->bBodyCrcError)
        bBCE = TRUE;
      if (pBlk->lBodyLengthError >= 0)
        bBLE = TRUE;
    }
  }

  if (bHCE || bBCE || bBLE)
    printf("\n");
  if (bHCE)
    printf("   HCE = Head CRC error\n");
  if (bBCE)
    printf("   BCE = Body CRC error\n");
  if (bBLE)
    printf("   BLE = Body Length error\n");
}
//---------------------------------------------------------------------------
int readTapeFile(WORK_PARAMS *pWP)
{
  FILE *f;
  TAPE_BLOCK *pLastBlk, blkTmp;
  long lPosH, lPosB = 0, lFileSize, lTemp;
  BOOL bHdr, bOldType;
  int iErr;
  WORD wTemp;

  printf("Reading source file: %s\n", pWP->pFnIn);

  if (pWP->pBlocks)
    freeAllBlocks(pWP->pBlocks);
  pWP->pBlocks = NULL;

  iErr = ERR_OK;
  do {
    f = fopen(pWP->pFnIn, "rb");
    if (f == NULL) {
      iErr = printError(ERR_OPEN_SRC_FILE, pWP->pFnIn, NULL);
      break;
    }
    if (fseek(f, 0L, SEEK_END) != 0) {
      iErr = printError(ERR_ACCESS_SRC_FILE, pWP->pFnIn, NULL);
      break;
    }
    lFileSize = ftell(f);
    if (lFileSize < 0) {
      iErr = printError(ERR_LENGTH_SRC_FILE, pWP->pFnIn, NULL);
      break;
    }
    if (fseek(f, 0L, SEEK_SET) != 0) {
      iErr = printError(ERR_ACCESS_SRC_FILE, pWP->pFnIn, NULL);
      break;
    }
    if (lFileSize < 65) {
      iErr = printError(ERR_TAPE_CORRUPTED, pWP->pFnIn, NULL);
      break;
    }

    bOldType = FALSE;
    do {
      memset(pWP->pbtBuffer, 0, 67);
      memset(&blkTmp, 0, sizeof(TAPE_BLOCK));
      bHdr = FALSE;

      lPosH = setFilePointer(f, 0, SEEK_CUR);
      if (lPosH < 0) {
        iErr = printError(ERR_ACCESS_SRC_FILE, pWP->pFnIn, NULL);
        break;
      }

      lTemp = fread(pWP->pbtBuffer, 1, 67, f);   // 2 + 63 + 2
      if (lTemp == 0) {
        iErr = printError(ERR_READ_SRC_FILE, pWP->pFnIn, NULL);
        break;
      }

      wTemp = *(WORD *) pWP->pbtBuffer;
      if (lTemp >= 63) {
        bHdr = checkHeader(pWP->pbtBuffer, &blkTmp);
        if (bHdr == TRUE) {
          bOldType = TRUE; // old tape pFS
          lPosB = lPosH + 63;
          wTemp = 0;
        }
        else if (wTemp == 63 && lTemp == 67) {
          bHdr = checkHeader(pWP->pbtBuffer + 2, &blkTmp);
          if (bHdr == TRUE) {
            lPosH += 2;                    // posun za dlzku bloku hlavicky
            lPosB = lPosH + 63 + 2;        // posun za hlavicku a dl. bloku
            wTemp = *(WORD *) (pWP->pbtBuffer + 65); // dlzka bloku za hlavickou
          }
        }
      }

      if (bHdr == TRUE) {
        lTemp = setFilePointer(f, lPosB, SEEK_SET);
        if ((long) (blkTmp.wLen + 2) > lFileSize - lTemp
            || (bOldType == FALSE && blkTmp.wLen + 2 != wTemp)) {
          if (bOldType == FALSE) {
            blkTmp.lBodyLengthError = wTemp - 2;
            lTemp = setFilePointer(f, wTemp, SEEK_CUR);
          }
          else {
            blkTmp.lBodyLengthError = lFileSize - lTemp - 2;
            lTemp = lFileSize;
          }
        }
        else {
          lTemp = fread(pWP->pbtBuffer, 1, blkTmp.wLen + 2, f);   // 2 + 63 + 2
          if (lTemp == 0) {
            iErr = printError(ERR_READ_SRC_FILE, pWP->pFnIn, NULL);
            break;
          }
          blkTmp.lBodyLengthError = -1;
          blkTmp.bBodyCrcError = !checkCrc(pWP->pbtBuffer, blkTmp.wLen + 1, &blkTmp.btBodyCrc);
          lTemp = setFilePointer(f, 0, SEEK_CUR);
        }

        blkTmp.lOffsetHeader = lPosH;
        blkTmp.lOffsetBody = lPosB;
      }
      else {
        if (bOldType == TRUE) {
          if ((lFileSize - lPosH) > 65535)
            break;

          blkTmp.wLen = (WORD) (lFileSize - lPosH);
          lTemp = lFileSize;
        }
        else {
          if ((long) wTemp > (lFileSize - lPosH))
            break;

          blkTmp.wLen = wTemp;
          lPosH += 2;
          lTemp = setFilePointer(f, lPosH + wTemp, SEEK_SET);
        }

        blkTmp.cTyp = 0;
        blkTmp.lOffsetBody = lPosH;
        blkTmp.lBodyLengthError = -1;
      }

      if (pWP->pBlocks == NULL) {
        iErr = allocBuffer((void *) &pWP->pBlocks, sizeof(TAPE_BLOCK));
        pLastBlk = pWP->pBlocks;
        pLastBlk->prev = NULL;
      }
      else {
        iErr = allocBuffer((void *) &pLastBlk->next, sizeof(TAPE_BLOCK));
        pLastBlk->next->prev = pLastBlk;
        pLastBlk = pLastBlk->next;
      }
      if (iErr < ERR_OK)
        break;

      memcpy(pLastBlk, &blkTmp, sizeof(TAPE_BLOCK) - 2 * sizeof(TAPE_BLOCK *));

      pLastBlk->bRawFile = FALSE;
      pLastBlk->next = NULL;

      pWP->iBlockCount++;

    } while (lTemp < lFileSize - 2);

    if (lTemp != lFileSize)
      iErr = printError(ERR_TAPE_CORRUPTED, pWP->pFnIn, &lTemp);

  } while (FALSE);

  if (f != NULL)
    fclose(f);

  if (iErr == ERR_OK && pWP->bVB)
    printf(" Blocks: %d\n", pWP->iBlockCount);

  return iErr;
}
//---------------------------------------------------------------------------
int loadTapeBlock(WORK_PARAMS *pWP, int iBlk, BOOL bHead, TAPE_BLOCK **ppBlk)
{
  TAPE_BLOCK *pBlk;
  int iErr;
  long lOffset;
  DWORD dwDataLen;

  iErr = findBlock(pWP, iBlk, &pBlk);
  if (iErr != ERR_OK)
    return iErr;

  *ppBlk = pBlk;
  if (bHead && pBlk->cTyp != 0) {
    lOffset = pBlk->lOffsetHeader;
    dwDataLen = 63;
    if (pWP->bMato) {
      lOffset += (3 * 16);
      dwDataLen -= (3 * 16);
    }
  }
  else {
    lOffset = pBlk->lOffsetBody;
    dwDataLen = pBlk->wLen;
    if (pBlk->cTyp != 0)
      dwDataLen += 2;
  }

  iErr = readFile(pWP->pFnIn, pWP->pbtBuffer, lOffset, dwDataLen);
  if (iErr != ERR_OK)
    return iErr;

  return (int) dwDataLen;
}
//---------------------------------------------------------------------------
int findBlock(WORK_PARAMS *pWP, int iBlk, TAPE_BLOCK **ppBlk)
{
  TAPE_BLOCK *pBlk;
  int iCnt;

  iCnt = 1;
  pBlk = pWP->pBlocks;
  while (pBlk) {
    if (iCnt == iBlk) {
      *ppBlk = pBlk;
      return ERR_OK;
    }

    pBlk = pBlk->next;
    iCnt++;
  }

  return printError(ERR_BAD_BLOCK_NUMBER, &iBlk, NULL);
}
//---------------------------------------------------------------------------
void freeAllBlocks(TAPE_BLOCK *pBlocks)
{
  TAPE_BLOCK *pBlk, *pNextBlk;

  pBlk = pBlocks;
  while (pBlk) {
    pNextBlk = pBlk->next;
    free(pBlk);
    pBlk = pNextBlk;
  }
}
//---------------------------------------------------------------------------
BOOL checkCrc(BYTE *pbtBuff, int iLength, BYTE *pGoodCrc)
{
  BYTE btCrc = 0;

  while (iLength-- > 0)
    btCrc += *pbtBuff++;

  if (pGoodCrc)
    *pGoodCrc = btCrc;

  return (btCrc == *pbtBuff);
}
//---------------------------------------------------------------------------
BOOL checkHeader(BYTE *pbtBuff, TAPE_BLOCK *pBlk)
{
  int ii;

  for (ii = 0; ii < 16; ii++) {
    if (pbtBuff[ii] != 0xFF)
      return FALSE;
  }

  for (ii = 16; ii < 32; ii++) {
    if (pbtBuff[ii] != 0x00)
      return FALSE;
  }

  for (ii = 32; ii < 48; ii++) {
    if (pbtBuff[ii] != 0x55)
      return FALSE;
  }

  memcpy(pBlk, pbtBuff + 48, 15);
  pBlk->bHeadCrcError = !checkCrc(pbtBuff + 48, 14, NULL);

  return TRUE;
}
//---------------------------------------------------------------------------

