//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "common.h"

//---------------------------------------------------------------------------
int writeSilentGap(WORK_PARAMS *pWP, int iSG);
int writeBlockData(WORK_PARAMS *pWP, int iSize);
int writeBlockPmd85(WORK_PARAMS *pWP, int iBlk);
int writeLeaderTonePmd85(WORK_PARAMS *pWP, double dDuration);
int writeBytePmd85(WORK_PARAMS *pWP, BYTE btByte);
int writeBlockMato(WORK_PARAMS *pWP, int iBlk);
int writeLeaderToneMato(WORK_PARAMS *pWP, int iPulses, BOOL bAddLog1);
int writeByteMato(WORK_PARAMS *pWP, BYTE btByte);
int prepareOutputFile(WORK_PARAMS *pWP);
int finalizeOutputFile(WORK_PARAMS *pWP);
int prepareLogsBuffer(WORK_PARAMS *pWP);
int prepareWaveHeader(WORK_PARAMS *pWP);

//---------------------------------------------------------------------------
int createWav(WORK_PARAMS *pWP)
{
  int iErr, iIdx, iBlk;

  printf("Protocol: ");
  if (pWP->bMato)
    printf("Mato");
  else
    printf("PMD 85");
  if (pWP->bDoubleSpeed || pWP->bRevPhase) {
    printf(" (");
    if (pWP->bDoubleSpeed)
      printf("Double Speed");
    if (pWP->bDoubleSpeed && pWP->bRevPhase)
      printf(", ");
    if (pWP->bRevPhase)
      printf("Reversal Phase");
    printf(")");
  }
  printf("\n");

  do {
    iErr = prepareOutputFile(pWP);
    if (iErr != ERR_OK)
      break;

    iErr = writeSilentGap(pWP, 1);
    if (iErr != ERR_OK)
      break;

    for (iIdx = 0; iIdx < pWP->iBlkListCnt; iIdx++) {
      iBlk = pWP->piBlkNumList[iIdx];
      if (pWP->bMato)
        iErr = writeBlockMato(pWP, iBlk);
      else
        iErr = writeBlockPmd85(pWP, iBlk);
      if (iErr != ERR_OK)
        break;
    }
    if (iErr != ERR_OK)
      break;

    iErr = writeSilentGap(pWP, 1);
    if (iErr != ERR_OK)
      break;

    iErr = finalizeOutputFile(pWP);
    if (iErr != ERR_OK)
      break;

  } while (FALSE);

  return ERR_OK;
}
//---------------------------------------------------------------------------
int writeSilentGap(WORK_PARAMS *pWP, int iSG)
{
  int iErr = ERR_OK;
  int iCount = 0, iPulses, iPulseLen;
  BYTE btSilent[64];

  if (pWP->bMato) {
    iPulses = (int) (iSG * PULSE_FREQ);
    iPulseLen = pWP->iLog0Length / 3;
  }
  else {
    iPulses = (int) (iSG * (pWP->bDoubleSpeed ? PULSE_FREQ_2400 : PULSE_FREQ_1200));
    iPulseLen = pWP->iLog0Length;
  }

  memset(btSilent, MI_STATE, iPulseLen);

  do {
    if (fwrite(btSilent, iPulseLen, 1, pWP->fOut) != 1) {
      iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
      break;
    }
  } while (++iCount < iPulses);

  return iErr;
}
//---------------------------------------------------------------------------
int writeBlockData(WORK_PARAMS *pWP, int iSize)
{
  int iErr = ERR_OK, i;
  BYTE *p = pWP->pbtBuffer;
  for (i = 0; i < iSize; i++) {
    if (pWP->bMato)
      iErr = writeByteMato(pWP, *p++);
    else
      iErr = writeBytePmd85(pWP, *p++);
    if (iErr != ERR_OK)
      break;
  }

  return iErr;
}
//---------------------------------------------------------------------------
int writeBlockPmd85(WORK_PARAMS *pWP, int iBlk)
{
  int iErr, iSize;
  TAPE_BLOCK *pBlk = NULL;
  char szName[9] = { 0 };

  do {
    iErr = loadTapeBlock(pWP, iBlk, TRUE, &pBlk);
    if (iErr < ERR_OK)
      break;
    iSize = iErr;

    if (pWP->bVB) {
      if (pBlk->cTyp != 0) {
        memcpy(szName, pBlk->cMeno, 8);
        printf("Exporting block %d: %02d/%c %8.8s\n",
               iBlk, (int) pBlk->btCislo, pBlk->cTyp, szName);
      }
      else
        printf("Exporting headerless block %d (%5d bytes)\n", iBlk, (int) pBlk->wLen);
    }

    if (iBlk != pWP->piBlkNumList[0] && pBlk->cTyp != 0) {
      iErr = writeLeaderTonePmd85(pWP, TAIL_DURATION);
      if (iErr != ERR_OK)
        break;

      iErr = writeSilentGap(pWP, pWP->iSilentGap);
      if (iErr != ERR_OK)
        break;
    }

    iErr = writeLeaderTonePmd85(pWP, (pBlk->cTyp != 0) ? HEAD_LEADER_DURATION : BODY_LEADER_DURATION);
    if (iErr != ERR_OK)
      break;

    iErr = writeBlockData(pWP, iSize);
    if (iErr != ERR_OK)
      break;

    if (pBlk->cTyp != 0) {
      iErr = loadTapeBlock(pWP, iBlk, FALSE, &pBlk);
      if (iErr < ERR_OK)
        break;
      iSize = iErr;

      iErr = writeLeaderTonePmd85(pWP, HBODY_LEADER_DURATION);
      if (iErr != ERR_OK)
        break;

      iErr = writeBlockData(pWP, iSize);
      if (iErr != ERR_OK)
        break;
    }

    if (iBlk == pWP->piBlkNumList[pWP->iBlkListCnt - 1]) {
      iErr = writeLeaderTonePmd85(pWP, TAIL_DURATION);
      if (iErr != ERR_OK)
        break;
    }

  } while (FALSE);

  return iErr;
}
//---------------------------------------------------------------------------
int writeLeaderTonePmd85(WORK_PARAMS *pWP, double dDuration)
{
  int iErr = ERR_OK;
  int iCount = 0;
  int iPulses = (int) (dDuration * (pWP->bDoubleSpeed ? PULSE_FREQ_2400 : PULSE_FREQ_1200));

  do {
    if (fwrite(pWP->pbtLog1, pWP->iLog1Length, 1, pWP->fOut) != 1) {
      iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
      break;
    }
  } while (++iCount < iPulses);

  return iErr;
}
//---------------------------------------------------------------------------
int writeBytePmd85(WORK_PARAMS *pWP, BYTE btByte)
{
  int iErr = ERR_OK;
  BYTE btRoll;
  int iBit;

  // Start bit
  if (fwrite(pWP->pbtLog0, pWP->iLog0Length, 1, pWP->fOut) != 1) {
    iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
    return iErr;
  }

  btRoll = 1;
  for (iBit = 0; iBit < 8; iBit++) {
    if (btByte & btRoll) {
      // Log.1
      if (fwrite(pWP->pbtLog1, pWP->iLog1Length, 1, pWP->fOut) != 1) {
        iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
        break;
      }
    }
    else {
      // Log.0
      if (fwrite(pWP->pbtLog0, pWP->iLog0Length, 1, pWP->fOut) != 1) {
        iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
        break;
      }
    }
    btRoll <<= 1;
  }

  if (iErr != ERR_OK)
    return iErr;

  // 1. Stop bit
  if (fwrite(pWP->pbtLog1, pWP->iLog1Length, 1, pWP->fOut) != 1) {
    iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
    return iErr;
  }

  // 2. Stop bit
  if (fwrite(pWP->pbtLog1, pWP->iLog1Length, 1, pWP->fOut) != 1)
    iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);

  return iErr;
}
//---------------------------------------------------------------------------
int writeBlockMato(WORK_PARAMS *pWP, int iBlk)
{
  int iErr, iSize;
  TAPE_BLOCK *pBlk = NULL;
  char szName[9] = { 0 };

  do {
    iErr = loadTapeBlock(pWP, iBlk, TRUE, &pBlk);
    if (iErr < ERR_OK)
      break;
    iSize = iErr;

    if (pWP->bVB) {
      if (pBlk->cTyp != 0) {
        memcpy(szName, pBlk->cMeno, 8);
        printf("Exporting block %d: %02d/%c %8.8s\n",
               iBlk, (int) pBlk->btCislo, pBlk->cTyp, szName);
      }
      else
        printf("Exporting headerless block %d (%5d bytes)\n", iBlk, (int) pBlk->wLen);
    }

    if (iBlk != pWP->piBlkNumList[0] && pBlk->cTyp != 0) {
      iErr = writeSilentGap(pWP, pWP->iSilentGap);
      if (iErr != ERR_OK)
        break;
    }

    iErr = writeLeaderToneMato(pWP, (pBlk->cTyp != 0) ? HEAD_LEADER_PULSES : BODY_LEADER_PULSES, FALSE);
    if (iErr != ERR_OK)
      break;
    if (pBlk->cTyp != 0) {
      iErr = writeByteMato(pWP, 0x55);
      if (iErr != ERR_OK)
        break;
    }

    iErr = writeBlockData(pWP, iSize);
    if (iErr != ERR_OK)
      break;

    if (pBlk->cTyp != 0) {
      iErr = loadTapeBlock(pWP, iBlk, FALSE, &pBlk);
      if (iErr < ERR_OK)
        break;
      iSize = iErr;

      iErr = writeLeaderToneMato(pWP, HBODY_LEADER_PULSES, TRUE);
      if (iErr != ERR_OK)
        break;

      iErr = writeBlockData(pWP, iSize);
      if (iErr != ERR_OK)
        break;
    }

  } while (FALSE);

  return iErr;
}
//---------------------------------------------------------------------------
int writeLeaderToneMato(WORK_PARAMS *pWP, int iPulses, BOOL bAddLog1)
{
  int iErr = ERR_OK;
  int iCount = 0;

  do {
    if (fwrite(pWP->pbtLog0, pWP->iLog0Length, 1, pWP->fOut) != 1) {
      iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
      break;
    }
  } while (++iCount < iPulses);

  if (bAddLog1) {
    // Log.1
    if (fwrite(pWP->pbtLog1, pWP->iLog1Length, 1, pWP->fOut) != 1) {
      iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
    }
  }

  return iErr;
}
//---------------------------------------------------------------------------
int writeByteMato(WORK_PARAMS *pWP, BYTE btByte)
{
  int iErr = ERR_OK;
  BYTE btRoll;
  int iBit;

  // Start bit
  if (fwrite(pWP->pbtLog1, pWP->iLog1Length, 1, pWP->fOut) != 1) {
    iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
    return iErr;
  }

  btRoll = 1;
  for (iBit = 0; iBit < 8; iBit++) {
    if (btByte & btRoll) {
      // Log.1
      if (fwrite(pWP->pbtLog1, pWP->iLog1Length, 1, pWP->fOut) != 1) {
        iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
        break;
      }
    }
    else {
      // Log.0
      if (fwrite(pWP->pbtLog0, pWP->iLog0Length, 1, pWP->fOut) != 1) {
        iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
        break;
      }
    }
    btRoll <<= 1;
  }

  return iErr;
}
//---------------------------------------------------------------------------
int prepareOutputFile(WORK_PARAMS *pWP)
{
  int iErr;

  printf("Creating output WAV file: %s\n", pWP->pFnOut);

  do {
    iErr = prepareLogsBuffer(pWP);
    if (iErr != ERR_OK)
      break;

    iErr = prepareWaveHeader(pWP);
    if (iErr != ERR_OK)
      break;

    pWP->fOut = fopen(pWP->pFnOut, "wb");
    if (pWP->fOut == NULL) {
      iErr = printError(ERR_OPEN_DEST_FILE, pWP->pFnOut, NULL);
      break;
    }

    if (fwrite(pWP->pWH, 1, sizeof(WAVE_HEADER), pWP->fOut) != sizeof(WAVE_HEADER)) {
      iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
      break;
    }
  } while (FALSE);

  if (pWP->bVB && iErr == ERR_OK)
    printf(" Sample rate: %d Hz\n", pWP->iSampleRate);

  return iErr;
}
//---------------------------------------------------------------------------
int finalizeOutputFile(WORK_PARAMS *pWP)
{
  int iErr = ERR_OK;
  long lOutLen, lOffset;

  do {
    lOutLen = ftell(pWP->fOut);
    if (lOutLen < 0) {
      iErr = printError(ERR_ACCESS_DEST_FILE, pWP->pFnOut, NULL);
      break;
    }

    lOffset = ((BYTE *) &pWP->pWH->dwTotLength) - ((BYTE *) pWP->pWH);
    if (fseek(pWP->fOut, lOffset, SEEK_SET) != 0) {
      iErr = printError(ERR_ACCESS_DEST_FILE, pWP->pFnOut, NULL);
      break;
    }

    lOutLen -= (sizeof(FOURCC) + sizeof(DWORD));
    if (fwrite(&lOutLen, 1, sizeof(DWORD), pWP->fOut) != sizeof(DWORD)) {
      iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
      break;
    }

    lOffset = ((BYTE *) &pWP->pWH->dwDataLength) - ((BYTE *) pWP->pWH);
    if (fseek(pWP->fOut, lOffset, SEEK_SET) != 0) {
      iErr = printError(ERR_ACCESS_DEST_FILE, pWP->pFnOut, NULL);
      break;
    }

    lOutLen -= (sizeof(WAVE_HEADER) - sizeof(FOURCC) - sizeof(DWORD));
    if (fwrite(&lOutLen, 1, sizeof(DWORD), pWP->fOut) != sizeof(DWORD)) {
      iErr = printError(ERR_WRITE_DEST_FILE, pWP->pFnOut, NULL);
      break;
    }

    fclose(pWP->fOut);
    pWP->fOut = NULL;

  } while (FALSE);

  if (iErr == ERR_OK && pWP->bVB)
    printf("Output WAV file finalized and closed.\n");

  return iErr;
}
//---------------------------------------------------------------------------
int prepareLogsBuffer(WORK_PARAMS *pWP)
{
  int iErr, iLevelsBuffSize, iOffStateOffset, iOffStateLen, iLog0Offset;
  BYTE btON, btOFF;

  if (pWP->bMato) {
    switch (pWP->iSampleRateType) {
      case SAMPLE_RATE_LOW :
        pWP->iSampleRate = SAMPLE_RATE_LOW_HZ;
        pWP->iLog0Length = 3 * PULSE_SIZE_L;
        pWP->iLog1Length = 5 * PULSE_SIZE_L;
        iLevelsBuffSize = 5 * PULSE_SIZE_L;
        iOffStateOffset = 3 * PULSE_SIZE_L;
        iOffStateLen = 2 * PULSE_SIZE_L;
        iLog0Offset = 2 * PULSE_SIZE_L;
        break;

      case SAMPLE_RATE_MIDDLE :
        pWP->iSampleRate = SAMPLE_RATE_MIDDLE_HZ;
        pWP->iLog0Length = 3 * PULSE_SIZE_M;
        pWP->iLog1Length = 5 * PULSE_SIZE_M;
        iLevelsBuffSize = 5 * PULSE_SIZE_M;
        iOffStateOffset = 3 * PULSE_SIZE_M;
        iOffStateLen = 2 * PULSE_SIZE_M;
        iLog0Offset = 2 * PULSE_SIZE_M;
        break;

      case SAMPLE_RATE_HIGH :
        pWP->iSampleRate = SAMPLE_RATE_HIGH_HZ;
        pWP->iLog0Length = 3 * PULSE_SIZE_H;
        pWP->iLog1Length = 5 * PULSE_SIZE_H;
        iLevelsBuffSize = 5 * PULSE_SIZE_H;
        iOffStateOffset = 3 * PULSE_SIZE_H;
        iOffStateLen = 2 * PULSE_SIZE_H;
        iLog0Offset = 2 * PULSE_SIZE_H;
        break;
    }
  }
  else {
    switch (pWP->iSampleRateType) {
      case SAMPLE_RATE_LOW :
        if (pWP->bDoubleSpeed) {
          pWP->iSampleRate = SAMPLE_RATE_LOW_HZ;
          pWP->iLog0Length = PULSE_SIZE_2400_L;
          pWP->iLog1Length = PULSE_SIZE_2400_L;
          iLevelsBuffSize = (PULSE_SIZE_2400_L / 2) * 3;
          iOffStateOffset = PULSE_SIZE_2400_L / 2;
          iOffStateLen = PULSE_SIZE_2400_L / 2;
          iLog0Offset = PULSE_SIZE_2400_L / 2;
        }
        else {
          pWP->iSampleRate = SAMPLE_RATE_LOW_HZ;
          pWP->iLog0Length = PULSE_SIZE_1200_L;
          pWP->iLog1Length = PULSE_SIZE_1200_L;
          iLevelsBuffSize = (PULSE_SIZE_1200_L / 2) * 3;
          iOffStateOffset = PULSE_SIZE_1200_L / 2;
          iOffStateLen = PULSE_SIZE_1200_L / 2;
          iLog0Offset = PULSE_SIZE_1200_L / 2;
        }
        break;

      case SAMPLE_RATE_MIDDLE :
        if (pWP->bDoubleSpeed) {
          pWP->iSampleRate = SAMPLE_RATE_MIDDLE_HZ;
          pWP->iLog0Length = PULSE_SIZE_2400_M;
          pWP->iLog1Length = PULSE_SIZE_2400_M;
          iLevelsBuffSize = (PULSE_SIZE_2400_M / 2) * 3;
          iOffStateOffset = PULSE_SIZE_2400_M / 2;
          iOffStateLen = PULSE_SIZE_2400_M / 2;
          iLog0Offset = PULSE_SIZE_2400_M / 2;
        }
        else {
          pWP->iSampleRate = SAMPLE_RATE_MIDDLE_HZ;
          pWP->iLog0Length = PULSE_SIZE_1200_M;
          pWP->iLog1Length = PULSE_SIZE_1200_M;
          iLevelsBuffSize = (PULSE_SIZE_1200_M / 2) * 3;
          iOffStateOffset = PULSE_SIZE_1200_M / 2;
          iOffStateLen = PULSE_SIZE_1200_M / 2;
          iLog0Offset = PULSE_SIZE_1200_M / 2;
        }
        break;

      case SAMPLE_RATE_HIGH :
        if (pWP->bDoubleSpeed) {
          pWP->iSampleRate = SAMPLE_RATE_HIGH_HZ;
          pWP->iLog0Length = PULSE_SIZE_2400_H;
          pWP->iLog1Length = PULSE_SIZE_2400_H;
          iLevelsBuffSize = (PULSE_SIZE_2400_H / 2) * 3;
          iOffStateOffset = PULSE_SIZE_2400_H / 2;
          iOffStateLen = PULSE_SIZE_2400_H / 2;
          iLog0Offset = PULSE_SIZE_2400_H / 2;
        }
        else {
          pWP->iSampleRate = SAMPLE_RATE_HIGH_HZ;
          pWP->iLog0Length = PULSE_SIZE_1200_H;
          pWP->iLog1Length = PULSE_SIZE_1200_H;
          iLevelsBuffSize = (PULSE_SIZE_1200_H / 2) * 3;
          iOffStateOffset = PULSE_SIZE_1200_H / 2;
          iOffStateLen = PULSE_SIZE_1200_H / 2;
          iLog0Offset = PULSE_SIZE_1200_H / 2;
        }
        break;
    }
  }

  iErr = allocBuffer((void *) &pWP->pbtLevels, iLevelsBuffSize);
  if (iErr != ERR_OK)
    return iErr;

  btON = (BYTE) ((pWP->bRevPhase) ? pWP->btAmpLo : pWP->btAmpHi);
  btOFF = (BYTE) ((pWP->bRevPhase) ? pWP->btAmpHi : pWP->btAmpLo);

  pWP->pbtLog1 = pWP->pbtLevels;
  pWP->pbtLog0 = pWP->pbtLevels + iLog0Offset;
  memset(pWP->pbtLog1, btON, iLevelsBuffSize);
  memset(pWP->pbtLog1 + iOffStateOffset, btOFF, iOffStateLen);

  return iErr;
}
//---------------------------------------------------------------------------
int prepareWaveHeader(WORK_PARAMS *pWP)
{
  int iErr = allocBuffer((void *) &pWP->pWH, sizeof(WAVE_HEADER));
  if (iErr != ERR_OK)
    return iErr;

  pWP->pWH->riff = MAKEFOURCC('R', 'I', 'F', 'F');
  pWP->pWH->dwTotLength = 0;
  pWP->pWH->wave = MAKEFOURCC('W', 'A', 'V', 'E');
  pWP->pWH->fmt = MAKEFOURCC('f', 'm', 't', ' ');
  pWP->pWH->dwHdrLength = sizeof(PCMWAVEFORMAT);
  pWP->pWH->waveFormat.wf.wFormatTag = WAVE_FORMAT_PCM;
  pWP->pWH->waveFormat.wf.nChannels = 1;
  pWP->pWH->waveFormat.wf.nSamplesPerSec = pWP->iSampleRate;
  pWP->pWH->waveFormat.wf.nAvgBytesPerSec = pWP->iSampleRate;
  pWP->pWH->waveFormat.wf.nBlockAlign = 1;
  pWP->pWH->waveFormat.wBitsPerSample = 8;
  pWP->pWH->data = MAKEFOURCC('d', 'a', 't', 'a');
  pWP->pWH->dwDataLength = 0;

  return ERR_OK;
}
//---------------------------------------------------------------------------

